/* m68hc05-tdep.c -- Motorola 68HC05 GDB Support
   Copyright 2005 Free Software Foundation, Inc.
   Copyright 2005 Martin Kgler
   Contributed by Martin Kgler (mkoegler@auto.tuwien.ac.at)

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA 02110-1301, USA.  */

#include "defs.h"
#include "frame.h"
#include "frame-unwind.h"
#include "frame-base.h"
#include "dwarf2-frame.h"
#include "trad-frame.h"
#include "symtab.h"
#include "gdbtypes.h"
#include "gdbcmd.h"
#include "gdbcore.h"
#include "gdb_string.h"
#include "value.h"
#include "inferior.h"
#include "dis-asm.h"
#include "symfile.h"
#include "objfiles.h"
#include "arch-utils.h"
#include "regcache.h"
#include "reggroups.h"

#include "target.h"
#include "elf-bfd.h"

struct gdbarch_tdep
{
  int elf_flags;
};

#define HARD_A_REGNUM        0
#define HARD_X_REGNUM        1
#define HARD_SP_REGNUM       2
#define HARD_PC_REGNUM       3
#define HARD_PSW_REGNUM      4
#define VIRTUAL_CARRY_REGNUM      5
#define M68HC05_NUM_REGS 6

#define SOFT_SP_REGNUM 7
#define SOFT_FP_REGNUM 8
#define M68HC05_SOFT_REGS 17

#define M68HC05_ALL_REGS (M68HC05_SOFT_REGS+M68HC05_NUM_REGS)
#define RET_REGISTER (M68HC05_NUM_REGS+4)

static char *m68hc05_register_names[] = {
  "a", "x", "hwsp", "pc", "ccr","carry",
  "AGP", "VFP", "SP", "FP",
  "RegB", "RegC", "RegD", "RegE", "RegF", "RegG",
  "RegH", "RegI", "RegJ", "RegK", "RegL", "RegM", "RegN"
};

struct m68hc05_soft_reg
{
  const char *name;
  CORE_ADDR addr;
};

static struct m68hc05_soft_reg soft_regs[M68HC05_ALL_REGS];

static int soft_min_addr;
static int soft_max_addr;
static int soft_reg_initialized = 0;
static int spbase = 0;

/* Look in the symbol table for the address of a pseudo register
   in memory.  If we don't find it, pretend the register is not used
   and not available.  */
static void
m68hc05_get_register_info (struct m68hc05_soft_reg *reg, const char *name)
{
  struct minimal_symbol *msymbol;

  msymbol = lookup_minimal_symbol (name, NULL, NULL);
  if (msymbol)
    {
      reg->addr = SYMBOL_VALUE_ADDRESS (msymbol);
      reg->name = xstrdup (name);

      /* Keep track of the address range for soft registers.  */
      if (reg->addr < (CORE_ADDR) soft_min_addr)
	soft_min_addr = reg->addr;
      if (reg->addr > (CORE_ADDR) soft_max_addr)
	soft_max_addr = reg->addr;
    }
  else
    {
      reg->name = 0;
      reg->addr = 0;
    }
}

/* Initialize the table of soft register addresses according
   to the symbol table.  */
static void
m68hc05_initialize_register_info (void)
{
  int i;
  struct minimal_symbol *msymbol;

  if (soft_reg_initialized)
    return;

  soft_min_addr = INT_MAX;
  soft_max_addr = 0;
  for (i = 0; i < M68HC05_ALL_REGS; i++)
    {
      soft_regs[i].name = 0;
    }

  msymbol = lookup_minimal_symbol ("SPBASE", NULL, NULL);
  if (msymbol)
    spbase = SYMBOL_VALUE_ADDRESS (msymbol);
  else
    {
      warning (_("No stack base found in the symbol table.\n"));
      warning (_("Stack backtrace will not work.\n"));
    }

  for (i = M68HC05_NUM_REGS; i < M68HC05_ALL_REGS; i++)
    m68hc05_get_register_info (&soft_regs[i], m68hc05_register_names[i]);

  if (soft_regs[SOFT_SP_REGNUM].name == 0)
    {
      warning (_("No stack soft register found in the symbol table.\n"));
      warning (_("Stack backtrace will not work.\n"));
    }
  soft_reg_initialized = 1;
}


static const char *
m68hc05_register_name (int reg_nr)
{
  if (reg_nr >= M68HC05_ALL_REGS)
    return NULL;

  return m68hc05_register_names[reg_nr];
}

/* Return the GDB type object for the "standard" data type
   of data in register N.  */

static struct type *
m68hc05_register_type (struct gdbarch *gdbarch, int reg_nr)
{
  switch (reg_nr)
    {
    case HARD_A_REGNUM:
    case HARD_X_REGNUM:
    case HARD_PSW_REGNUM:
    case VIRTUAL_CARRY_REGNUM:
      return builtin_type_uint8;

    case HARD_PC_REGNUM:
    case HARD_SP_REGNUM:
    case SOFT_SP_REGNUM:
    case SOFT_FP_REGNUM:
      return builtin_type_uint16;

    default:
      return builtin_type_uint8;
    }
}

static void
m68hc05_pseudo_register_read (struct gdbarch *gdbarch,
			      struct regcache *regcache, int regno, gdb_byte *buf)
{
  int size = 1;
  ULONGEST val;
  m68hc05_initialize_register_info ();

  /* Fetch a soft register: translate into a memory read.  */
  if (soft_regs[regno].name)
    {
      target_read_memory (soft_regs[regno].addr, buf, size);
    }
  else
    {
      memset (buf, 0, size);
    }

  if (regno == SOFT_SP_REGNUM || regno == SOFT_FP_REGNUM)
    {
      val = (*(unsigned char *) buf) + spbase;
      store_unsigned_integer (buf, 2, val);
    }
}

/* Store a pseudo register.  Translate the register store
   into a memory write.  */
static void
m68hc05_pseudo_register_write (struct gdbarch *gdbarch,
			       struct regcache *regcache,
			       int regno, const gdb_byte *buf)
{
  unsigned char val;
  m68hc05_initialize_register_info ();

  if (regno == SOFT_FP_REGNUM || regno == SOFT_SP_REGNUM)
    {
      val = extract_unsigned_integer (buf, 2) - spbase;
      buf = &val;
    }

  /* Store a soft register: translate into a memory write.  */
  if (soft_regs[regno].name)
    {
      const int regsize = 1;
      gdb_byte *tmp = alloca (regsize);
      memcpy (tmp, buf, regsize);
      target_write_memory (soft_regs[regno].addr, tmp, regsize);
    }
}

struct m68hc05_unwind_cache
{
  /* The previous frame's inner most stack address.  Used as this
     frame ID's stack_addr.  */
  CORE_ADDR prev_sp;
  CORE_ADDR sp;
  CORE_ADDR pc;

  struct trad_frame_saved_reg *saved_regs;

};


static CORE_ADDR
m68hc05_scan_prologue (CORE_ADDR pc, CORE_ADDR current_pc,
		       struct m68hc05_unwind_cache *info)
{
  return pc;
}

static CORE_ADDR
m68hc05_skip_prologue (CORE_ADDR pc)
{
  CORE_ADDR func_addr, func_end;
  struct symtab_and_line sal;
  struct m68hc05_unwind_cache tmp_cache = { 0 };

  /* If we have line debugging information, then the end of the
     prologue should be the first assembly instruction of the
     first source line.  */
  if (find_pc_partial_function (pc, NULL, &func_addr, &func_end))
    {
      sal = find_pc_line (func_addr, 0);
      if (sal.end && sal.end < func_end)
	return sal.end;
    }

  pc = m68hc05_scan_prologue (pc, (CORE_ADDR) - 1, &tmp_cache);
  return pc;
}

static const unsigned char *
m68hc05_breakpoint_from_pc (CORE_ADDR * pcptr, int *lenptr)
{
  static unsigned char breakpoint[] = { 0xA7 };

  *lenptr = sizeof (breakpoint);
  return breakpoint;
}

static int
gdb_print_insn_m68hc05 (bfd_vma memaddr, disassemble_info * info)
{
  return print_insn_m68hc05 (memaddr, info);
}


static CORE_ADDR
m68hc05_unwind_sp (struct gdbarch *gdbarch, struct frame_info *next_frame)
{
  ULONGEST sp;
  frame_unwind_unsigned_register (next_frame, HARD_SP_REGNUM, &sp);
  return sp;
}

static CORE_ADDR
m68hc05_unwind_pc (struct gdbarch *gdbarch, struct frame_info *next_frame)
{
  ULONGEST pc;

  frame_unwind_unsigned_register (next_frame, HARD_PC_REGNUM, &pc);
  return pc;
}

static struct frame_id
m68hc05_unwind_dummy_id (struct gdbarch *gdbarch,
			 struct frame_info *next_frame)
{
  ULONGEST tos;
  CORE_ADDR pc = frame_pc_unwind (next_frame);

  frame_unwind_unsigned_register (next_frame, HARD_SP_REGNUM, &tos);
  tos += 2;

  return frame_id_build (tos, pc);
}

struct m68hc05_unwind_cache *
m68hc05_frame_unwind_cache (struct frame_info *next_frame,
			    void **this_prologue_cache)
{
  ULONGEST prev_sp;
  ULONGEST base;
  struct m68hc05_unwind_cache *info;
  CORE_ADDR pc, current_pc;
  int i;
  struct symbol *s;

  if ((*this_prologue_cache))
    return (*this_prologue_cache);

  info = FRAME_OBSTACK_ZALLOC (struct m68hc05_unwind_cache);
  (*this_prologue_cache) = info;
  info->saved_regs = trad_frame_alloc_saved_regs (next_frame);

  frame_unwind_unsigned_register (next_frame, HARD_SP_REGNUM, &base);
  pc = frame_func_unwind (next_frame);

  current_pc = frame_pc_unwind (next_frame);
  if (info->pc != 0)
    m68hc05_scan_prologue (info->pc, current_pc, info);

  info->sp = base;
  info->prev_sp = base + 2;
  info->pc = pc;

  trad_frame_set_value (info->saved_regs, HARD_SP_REGNUM, info->prev_sp);
  if (info->sp < 0xfe)
    info->saved_regs[HARD_PC_REGNUM].addr = info->sp + 1;

  return info;
}


static void
m68hc05_frame_this_id (struct frame_info *next_frame,
		       void **this_prologue_cache, struct frame_id *this_id)
{
  struct m68hc05_unwind_cache *info
    = m68hc05_frame_unwind_cache (next_frame, this_prologue_cache);
  CORE_ADDR base;
  CORE_ADDR func;
  struct frame_id id;

  /* The FUNC is easy.  */
  func = frame_func_unwind (next_frame);

  /* Hopefully the prologue analysis either correctly determined the
     frame's base (which is the SP from the previous frame), or set
     that base to "NULL".  */
  base = info->prev_sp;
  if (base > 0xff)
    return;

  id = frame_id_build (base, func);
  (*this_id) = id;
}

static void
m68hc05_frame_prev_register (struct frame_info *next_frame,
			     void **this_prologue_cache,
			     int regnum, int *optimizedp,
			     enum lval_type *lvalp, CORE_ADDR * addrp,
			     int *realnump, void *bufferp)
{
  struct m68hc05_unwind_cache *info
    = m68hc05_frame_unwind_cache (next_frame, this_prologue_cache);

  trad_frame_get_prev_register (next_frame, info->saved_regs, regnum,
				optimizedp, lvalp, addrp, realnump, bufferp);

}


static const struct frame_unwind m68hc05_frame_unwind = {
  NORMAL_FRAME,
  m68hc05_frame_this_id,
  m68hc05_frame_prev_register
};

const struct frame_unwind *
m68hc05_frame_sniffer (struct frame_info *next_frame)
{
  return &m68hc05_frame_unwind;
}

/* Given a return value in `regbuf' with a type `valtype',
   extract and copy its value into `valbuf'.

   Return values are always passed via registers r25:r24:...  */

static void
m68hc05_extract_return_value (struct type *type, struct regcache *regcache,
                          gdb_byte *valbuf)
{
  ULONGEST c;
  int len;
  if (TYPE_LENGTH (type) == 1)
    {
      regcache_cooked_read_unsigned (regcache, RET_REGISTER+1, &c);
      store_unsigned_integer (valbuf, 1, c);
    }
  else
    {
      int i;
      for (i=0; i< TYPE_LENGTH (type); i++)
        {
          regcache_cooked_read (regcache, RET_REGISTER + i,
                                (bfd_byte *) valbuf + i);
        }
    }
}


static struct gdbarch *
m68hc05_gdbarch_init (struct gdbarch_info info, struct gdbarch_list *arches)
{
  struct gdbarch *gdbarch;
  struct gdbarch_tdep *tdep;
  int elf_flags;

  /* Extract the elf_flags if available.  */
  if (info.abfd != NULL
      && bfd_get_flavour (info.abfd) == bfd_target_elf_flavour)
    elf_flags = elf_elfheader (info.abfd)->e_flags;
  else
    elf_flags = 0;

  /* try to find a pre-existing architecture */
  for (arches = gdbarch_list_lookup_by_info (arches, &info);
       arches != NULL;
       arches = gdbarch_list_lookup_by_info (arches->next, &info))
    {
      if (gdbarch_tdep (arches->gdbarch)->elf_flags != elf_flags)
	continue;

      return arches->gdbarch;
    }

  /* Need a new architecture. Fill in a target specific vector.  */
  tdep = (struct gdbarch_tdep *) xmalloc (sizeof (struct gdbarch_tdep));
  gdbarch = gdbarch_alloc (&info, tdep);
  tdep->elf_flags = elf_flags;

  set_gdbarch_num_regs (gdbarch, M68HC05_NUM_REGS);
  set_gdbarch_num_pseudo_regs (gdbarch, M68HC05_SOFT_REGS);
  set_gdbarch_pseudo_register_read (gdbarch, m68hc05_pseudo_register_read);
  set_gdbarch_pseudo_register_write (gdbarch, m68hc05_pseudo_register_write);
  set_gdbarch_register_name (gdbarch, m68hc05_register_name);
  set_gdbarch_register_type (gdbarch, m68hc05_register_type);
  set_gdbarch_addr_bit (gdbarch, 16);
  set_gdbarch_pc_regnum (gdbarch, HARD_PC_REGNUM);
  set_gdbarch_print_insn (gdbarch, gdb_print_insn_m68hc05);
  set_gdbarch_skip_prologue (gdbarch, m68hc05_skip_prologue);
  set_gdbarch_inner_than (gdbarch, core_addr_lessthan);
  set_gdbarch_breakpoint_from_pc (gdbarch, m68hc05_breakpoint_from_pc);
  set_gdbarch_write_pc (gdbarch, generic_target_write_pc);
  set_gdbarch_unwind_pc (gdbarch, m68hc05_unwind_pc);
  set_gdbarch_unwind_sp (gdbarch, m68hc05_unwind_sp);
  set_gdbarch_extract_return_value (gdbarch, m68hc05_extract_return_value);

  set_gdbarch_unwind_dummy_id (gdbarch, m68hc05_unwind_dummy_id);
  frame_unwind_append_sniffer (gdbarch, dwarf2_frame_sniffer);
  frame_unwind_append_sniffer (gdbarch, m68hc05_frame_sniffer);

  set_gdbarch_short_bit (gdbarch, 16);
  set_gdbarch_int_bit (gdbarch, 16);
  set_gdbarch_float_bit (gdbarch, 32);
  set_gdbarch_double_bit (gdbarch, 64);
  set_gdbarch_long_double_bit (gdbarch, 64);
  set_gdbarch_long_bit (gdbarch, 32);
  set_gdbarch_ptr_bit (gdbarch, 16);
  set_gdbarch_long_long_bit (gdbarch, 64);

  return gdbarch;
}

extern initialize_file_ftype _initialize_m68hc05_tdep;	/* -Wmissing-prototypes */

void
_initialize_m68hc05_tdep (void)
{
  register_gdbarch_init (bfd_arch_m68hc05, m68hc05_gdbarch_init);

}
